(* Write .mli for widgets *)
open Tables
open Compile

let write_create_p :w wname =
  (w "val create : parent: 'a widget -> ?name: string ->\n" : unit); 
  begin
    try 
      let option = Hashtbl.find types_table key:"options" in
      let classdefs = List.assoc wname in:option.subtypes in
      let tklabels = List.map fun:gettklabel classdefs in
      let l = List.map classdefs fun:
      	begin fun fc ->
     	  begin let p = gettklabel fc in
      	    if count p in:tklabels > 1 then small fc.ml_name else p
      	  end, fc.template
	end in
      w (catenate_sep sep:" ->\n" 
      	 (List.map l fun:
	  begin fun (s,t) ->
	    "  ?" ^ s ^ ":"
	    ^(ppMLtype
 	     (match types_of_template t with
		[t] -> labeloff t at:"write_create_p"
	     | [] -> fatal_error "multiple"
	     | l -> Product (List.map fun:(labeloff at:"write_create_p") l)))
	  end))
    with Not_found -> fatal_error "in write_create_p"
  end;
  w (" -> "^wname^" widget\n");
  w "             (* [create p options ?name] creates a new widget with\n";
  w "                parent p and new patch component name.\n";
  w "                Options are restricted to the widget class subset,\n";
  w "                and checked dynamically. *)\n"

(* Unsafe: write special comment *)
let write_function_type :w def =
  if not def.safe then w "(* unsafe *)\n";
  w "val "; w def.ml_name; w " : ";
  let ts, os =
    let tys = types_of_template def.template in
    let rec replace_args l =
      match l with
    	    [] -> [], []
      | ((_,(List(Subtype(_,_)))) as x)::ls -> 
    	  let l, o = replace_args ls in
    	  l, x::o
      | ((l,_) as x)::ls when (try l.[0] with _ -> ' ') = '?' ->
      	  let l, o = replace_args ls in
	  l, x::o
      | x::ls ->
    	  let l, o = replace_args ls in
    	  x::l, o
    in
      replace_args tys
  in
  begin match ts, os with
    [],[] -> w "unit ->"
  | ls,os ->
      if ls = [] then w "?unit -> ";
      let counter = ref 0 in
      List.iter (ls @ os)
	fun:(fun (l,t) -> labelprint :w l; w (ppMLtype t :counter); w " -> ") 
  end;
  w (ppMLtype any:true return:true def.result); (* RETURN TYPE !!! *)
  w " \n";
(*  w "(* tk invocation: "; w (doc_of_template def.template); w " *)"; *)
  if def.safe then w "\n\n"
  else w "\n(* /unsafe *)\n\n"

let write_external_type :w def =
  match def.template with
    StringArg fname ->
      let ic = open_in_bin file:(fname ^ ".mli") in
      if not def.safe then w "(* unsafe *)\n";
      begin try
	while true do
	  w (input_line ic);
	  w "\n"
	done
      with End_of_file -> 
      	close_in ic;
	if def.safe then w "\n\n"
	else w "\n(* /unsafe *)\n\n"
      end
  | _ -> raise (Compiler_Error "invalid external definition")
